#ifdef WIN32
#include <process.h>
#include <shlobj.h>
#include <windows.h>
#else
#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <pthread.h>
#include <fcntl.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#ifdef IAO_SSL
#include "cyassl/openssl/ssl.h"
#include "cyassl/openssl/rand.h"
#include "cyassl/openssl/err.h"
#endif

#ifndef THREADS_H
#define THREADS_H

typedef void *iao_mutex;
typedef void *iao_thread;

iao_thread iao_spawnthread(void (*start_routine)(void*), void *arg);
iao_thread iao_spawnthread_nojoin(void (*start_routine)(void*), void *arg);

/* must call join thread if created by iao_spawnthread */
int iao_jointhread(iao_thread thread);
iao_mutex iao_createmutex();
int iao_lockmutex(iao_mutex mutex);
int iao_releasemutex(iao_mutex mutex);
int iao_deletemutex(iao_mutex mutex);
int iao_safecreatemutex(iao_mutex *mutex);

/* If you are calling the library from a thread not created by iao_threadcreate */
/* you must call iao_createglobalmutex before any other call to the ciao lib    */
void iao_createglobalmutex();


void iao_logmsg(char *msg);
void iao_logsysmsg(char *msg);

#endif
/*
** Copyright R-U-ON, 2006
** www.r-u-on.com
*/

#ifndef IAO_COMM_H
#define IAO_COMM_H

void iao_logmsg(char *msg);
void iao_logsysmsg(char *msg);

#define ERR_COMM_OKAY			 0
#define ERR_COMM_RESOURCE 		-1
#define ERR_COMM_HOSTNAME 		-2
#define ERR_COMM_NO_CONNECTION	-3
#define ERR_COMM_REFUSED		-4
#define ERR_COMM_TIMEOUT 		-5
#define ERR_COMM_DISCONNECTED 	-6
#define ERR_SSL_FAILED			-7
#define ERR_COMM_PORTINUSE      -8

typedef struct 
{
	int fd;
	char myhost[1024];
	char myip[20];
	char remotehost[1024];
	char remoteip[1024];
	int remoteport;
    	int myport;
	int err;
#ifdef IAO_SSL
	void *ssl;
#endif
	long udp;
} iao_connection;

int iao_tcpinit(void);
char *iao_comm_errstr(iao_connection *connection);

long iao_gethostbyname(char *host, char *ip);
int iao_connect(char *host, int port, iao_connection *result);
int iao_server(int port, iao_connection *result);
int iao_accept(iao_connection *server, iao_connection *client, int milisec);

int iao_udp(iao_connection *udp, int myport, char *host, int port);


#ifdef IAO_SSL
int iao_sslconnect(char *host, int port, iao_connection *result, int milisec);
int iao_sslify(iao_connection *result, int milisec);
#endif
int iao_send(iao_connection *s, char *buff, int len, int milisec);
int iao_recv(iao_connection *s, char *buff, int len, int milisec);
void iao_bridge(iao_connection *c1, iao_connection *c2, int millisec);
void iao_close(iao_connection *s);

#endif
/*
** Copyright R-U-ON, 2006
** www.r-u-on.com
*/

#ifndef IAO_HTTP_H
#define IAO_HTTP_H


struct iao_http_channel_struct
{
	iao_connection connection;
	int statuscode;
	char status[512];
	char contenttype[50];
	char realhost[512]; /* In case of proxy */
	int realport; /* In case of proxy */
	char proxycridentials[512]; /* In case of proxy */
};
typedef struct iao_http_channel_struct iao_http_channel;

int iao_geturl(iao_http_channel *channel, char *url, char **data);
int iao_geturlproxy(iao_http_channel *channel, char *url, char **data,
    char *proxyhost, int proxyport, char *proxycridentials);


int iao_http_open(char *host, int port, int secure, iao_http_channel *channel);
int iao_http_proxy_open(char *host, int port, int secure, char *proxyhost,
          int proxyport, char *proxycridetials, iao_http_channel *channel);
int iao_http_get(iao_http_channel* channel, char *url);
int iao_http_post(iao_http_channel* channel, char *url, char *data, int len);
int iao_http_getpage(iao_http_channel* channel, char **buff);
void iao_http_close(iao_http_channel* channel);
char *iao_http_status(int code);




#endif
/*
** Copyright R-U-ON, 2006
** www.r-u-on.com
*/

#ifndef IAO_BASE64_H
#define IAO_BASE64_H

int iao_base64encode(char *dest, void *src, int len);
int iao_base64decode(void *dest, char *src);

#endif
/*
** Copyright R-U-ON, 2006
** www.r-u-on.com
*/

#ifndef IAO_CONFIG_H
#define IAO_CONFIG_H

void iao_configinit();
int iao_parseconfig(char *buff, int len);

#endif
#include "ciao.h"



void *funcandarg(void *f, void *arg)
{
	void **farg = (void**)malloc(sizeof(void*)*2);
	farg[0] = f;
	farg[1] = arg;
	return farg;
}

void callfarg(void **farg)
{
	void (*threadroutine)(void*) = (void(*)(void*)) farg[0];
	void *arg = farg[1];
	free(farg);
	threadroutine(arg);
}

#ifndef WIN32

static  void *threadlauncher(void *arg)
{
	callfarg(arg);
	return NULL;
}

iao_thread iao_spawnthread(void (*threadroutine)(void*), void *arg)
{
	pthread_t t;
	void **farg = funcandarg((void*)threadroutine, arg);

	if (pthread_create(&t, NULL, threadlauncher, farg)==0)
	{
		return (iao_thread) t;
	}
	else
	{
		iao_logsysmsg("iao_spawnthread: Failed creating thread");
		return NULL;
	}
}
iao_thread iao_spawnthread_nojoin(void (*threadroutine)(void*), void *arg)
{
	return iao_spawnthread(threadroutine, arg);
}

int iao_jointhread(iao_thread thread)
{
	void *ret;
	if (pthread_join((pthread_t)thread, &ret)==0)
	{
		return 0;
	}
	else
	{
		iao_logsysmsg("iao_jointhread: Failed joining thread");
		return -1;
	}
}


iao_mutex iao_createmutex()
{
	pthread_mutex_t *mutex =
		 (pthread_mutex_t*) malloc(sizeof(pthread_mutex_t));

	if (pthread_mutex_init(mutex, NULL)==0)
	{
		return (iao_mutex) mutex;
	}
	else
	{
		iao_logsysmsg("iao_createmutex: Failed creating mutex");
		return NULL;
	}
}

int iao_lockmutex(iao_mutex mutex)
{
	if (pthread_mutex_lock( (pthread_mutex_t *) mutex)==0)
	{
		return 0;
	}
	else
	{
		iao_logsysmsg("iao_lockmutex: Failed locking mutex");
		return -1;
	}
}

int iao_releasemutex(iao_mutex mutex)
{
	if (pthread_mutex_unlock( (pthread_mutex_t*) mutex)==0)
	{
		return 0;
	}
	else
	{
		iao_logsysmsg("iao_releasemutex: Failed releasing mutex");
		return -1;
	}
}

int iao_deletemutex(iao_mutex mutex)
{
	if (pthread_mutex_destroy( (pthread_mutex_t*) mutex)==0)
	{
		free(mutex);
		return 0;
	}
	else
	{
		iao_logsysmsg("iao_deletemutex: Failed destroying mutex");
		return -1;
	}
}


#else

static  DWORD WINAPI threadlauncher(void *arg)
{
	callfarg(arg);
	return 0;
}

iao_thread iao_spawnthread(void (*threadroutine)(void*), void *arg)
{
	HANDLE t;

	void **farg = funcandarg(threadroutine, arg);

	t = CreateThread(NULL, 0,  threadlauncher, farg, 0, NULL);
	if (t!=NULL)
	{
		return (iao_thread) t;
	}
	else
	{
		iao_logsysmsg("iao_spawnthread: Failed creating thread");
		return NULL;
	}
}

iao_thread iao_spawnthread_nojoin(void (*threadroutine)(void*), void *arg)
{
	iao_thread thread = iao_spawnthread(threadroutine, arg);
	if (thread!=NULL)
	{
		CloseHandle((HANDLE)thread);
	}	
	return thread;
}

int iao_jointhread(iao_thread thread)
{
	void *ret = NULL;
	if (WaitForSingleObject((HANDLE)thread, INFINITE)==WAIT_OBJECT_0)
	{
		CloseHandle((HANDLE)thread);
		return 0;
	}
	else
	{
		iao_logsysmsg("Failed joining thread");
		return -1;
	}
}


iao_mutex iao_createmutex()
{
	HANDLE mutex = CreateMutex(NULL, FALSE, NULL);
	if (mutex!=NULL)
	{
		return (iao_mutex) mutex;
	}
	else
	{
		iao_logsysmsg("Failed creating mutex");
		return NULL;
	}
}

int iao_lockmutex(iao_mutex mutex)
{
	if (WaitForSingleObject((HANDLE)mutex, INFINITE)==WAIT_OBJECT_0)
	{
		return 0;
	}
	else
	{
		iao_logsysmsg("Failed locking mutex");
		return -1;
	}
}

int iao_releasemutex(iao_mutex mutex)
{
	if (ReleaseMutex((HANDLE) mutex)==TRUE)
	{
		return 0;
	}
	else
	{
		iao_logsysmsg("Failed releasing mutex");
		return -1;
	}
}

int iao_deletemutex(iao_mutex mutex)
{
	if (CloseHandle((HANDLE)mutex)==TRUE)
	{
		return 0;
	}
	else
	{
		iao_logsysmsg("Failed destroying mutex");
		return -1;
	}
}

#endif /* WIN32 */



#ifdef THREAD_TEST
char *agentname() { return "TEST"; }
void *mutex;
int counter = 10000000;

void thread(void *arg)
{
	int iter = 0;

	int *var = (int*) arg;
	while (*var>0)
	{
		iao_lockmutex(mutex);
		*var = *var - 1;
		iao_releasemutex(mutex);
		iter++;
	}
	printf("iter: %d\n", iter);
}

void *t[3];

int main()
{
	int i;
	mutex = iao_createmutex();
	for (i=0;i<3;++i)
	{
		t[i] = iao_spawnthread(thread, &counter);
	}
	for (i=0;i<3;++i)
	{
		int ret = iao_jointhread(t[i]);
		printf("%d\n", ret);
	}
	iao_deletemutex(mutex);
	return 0;
}
#endif
/*
** Copyright R-U-ON, 2006
** www.r-u-on.com
*/

#ifdef WIN32
#pragma warning(disable : 4996)
#else
#endif


#ifdef IAO_SSL
static SSL_CTX *sslctx = NULL;
#endif

static iao_mutex ghbn_mutex = NULL;


static struct timeval gettimeval(int milisec)
{
   struct timeval t;
   t.tv_sec=milisec / 1000;
   t.tv_usec=(milisec % 1000) * 1000;
   return t;
}

static void iptostr(unsigned long ip, char *buff)
{	
	sprintf(buff, "%d.%d.%d.%d",
		((unsigned char *)(&ip))[0],
		((unsigned char *)(&ip))[1],
		((unsigned char *)(&ip))[2],
		((unsigned char *)(&ip))[3]
		);
}


static int readyto(int send, iao_connection *s, int milisec)
{
	/* Variables */
	struct timeval time_struct,*tm;
	int err;
	int fd = s->fd;

	/* Mask */
	fd_set tmp_Mask, tmp_ExceptMask;
	FD_ZERO(&tmp_Mask);
	FD_SET(fd,&tmp_Mask);
	FD_ZERO(&tmp_ExceptMask);
	FD_SET(fd,&tmp_ExceptMask);

	/* time */
	if (milisec==-1)
	{
		tm=NULL;
	}
	else
	{
		tm=&time_struct;
		time_struct=gettimeval(milisec);
	}

	/* and go! */
	if (send)
	{
		err = select( (unsigned int) fd+1,0,&tmp_Mask,&tmp_ExceptMask,tm);
	}
	else
	{
		err = select( (unsigned int) fd+1,&tmp_Mask,0,&tmp_ExceptMask,tm);
	}
	if (err<0)
	{
		iao_logsysmsg(send ? "readytosend: select returned an error" : "readytorecv: select returned an error");
		s->err = ERR_COMM_RESOURCE;
		return -1;
	}

	if (FD_ISSET(fd,&tmp_ExceptMask))
	{
		iao_logsysmsg(send ? "readytosend: disconnected [exception bit]" : "readytorecv: disconnected [exception bit]");
		s->err = ERR_COMM_DISCONNECTED;
		return -1;
	}
	if (!(FD_ISSET(fd,&tmp_Mask)))
	{
		iao_logmsg(send ? "readytosend: timeout" : "readytorecv: timeout");
		s->err = ERR_COMM_TIMEOUT;
		return -1;
	}
	return 0;
}

static int readytosend(iao_connection *s, int milisec)
{
	return readyto(1, s, milisec);
}

static int readytorecv(iao_connection *s, int milisec)
{
	return readyto(0, s, milisec);
}

long iao_gethostbyname(char *host, char *ip)
{
        struct hostent *hp;
	long ret = 0;
	static long ruonipcache = 0;
	
	iao_lockmutex(ghbn_mutex);
	hp=gethostbyname(host);
        if (hp!=NULL)
        {
		ret = *(long *)hp->h_addr;
	}
	if (strcmp(host, "agent.r-u-on.com")==0) /* if you want to monitor a DNS... */
	{
		if (ret==0)
		{
			ret = ruonipcache;
		}
		else
		{
			ruonipcache = ret;
		}
	}
	if (ip!=NULL)
	{
		iptostr(ret, ip);
	}
	iao_releasemutex(ghbn_mutex);
	return ret;
}

static void resetconnectstruct(iao_connection *connection)
{
	connection->fd = -1;
	connection->myhost[0]=0;
	connection->myip[0]=0;
	connection->remotehost[0]=0;
	connection->remoteip[0]=0;
	connection->err=ERR_COMM_OKAY;
    connection->myport = 0;
    connection->remoteport = 0;
#ifdef IAO_SSL
	connection->ssl = NULL;
#endif
	connection->udp = 0;
}
    
int iao_connect(char *host, int port, iao_connection *result)
{
	/* variables */
	struct in_addr i_addr;
	struct sockaddr_in client;
	struct sockaddr_in addr;
	int keepalive_flag=1;
	unsigned int addrlen = sizeof(struct sockaddr_in);

	iao_tcpinit(); /* has an effect only the first time */

	/* reset struct */
    resetconnectstruct(result);

	/* Host to ip */
	i_addr.s_addr = iao_gethostbyname(host, result->remoteip);
	if (i_addr.s_addr==0)
	{
		iao_logsysmsg("iao_connect: Can't translate host to ip");
		result->err=ERR_COMM_HOSTNAME;
		return -1;
	}

	strcpy(result->remotehost , host);


	/* make struct */
	client.sin_addr = i_addr;
	client.sin_family = AF_INET;                
	client.sin_port   = htons(port);           

	/* Get a socket */
	if ((result->fd = (int)socket(AF_INET,SOCK_STREAM,0)) == -1) 
	{
		iao_logsysmsg("iao_connect: Can't obtain file descriptor");
		result->err=ERR_COMM_RESOURCE;
		return -1;
	}

	/* Set keep alive */
	if (setsockopt(result->fd, SOL_SOCKET, SO_KEEPALIVE,
		(char *) &keepalive_flag, sizeof(keepalive_flag)) == -1)
	{
		iao_logsysmsg("iao_connect: Set keep alive on socket failed");
		result->err=ERR_COMM_RESOURCE;
		iao_close(result);
		return -1;
	}

	/* connect */
	if (connect( result->fd, (struct sockaddr *)&client, sizeof(client)) == -1)
	{
#ifdef WIN32
		int e = WSAGetLastError();
		WSASetLastError(e);
		result->err =(e==WSAECONNREFUSED?ERR_COMM_REFUSED:ERR_COMM_NO_CONNECTION);
#else
		result->err=(errno==ECONNREFUSED ? ERR_COMM_REFUSED : ERR_COMM_NO_CONNECTION);
#endif
		iao_logsysmsg("iao_connect: Remote host unreachable");
		iao_close(result);
		return -1;
	}

	if (getsockname(result->fd, (struct sockaddr *)&addr, &addrlen) != -1) 
	{
		iptostr(addr.sin_addr.s_addr, result->myip);
	}

	result->remoteport = port;

	if (gethostname(result->myhost,sizeof(result->myhost))!=0) 
	{
		strcpy(result->myhost, "*undetermined*");
	}
	return 0;
}

#ifdef IAO_SSL

static SSL_SESSION *agentsslsession = NULL;
static int doreuse(iao_connection *c) 
{
	return strcmp(c->remotehost, "agent.r-u-on.com")==0 || strcmp(c->remotehost, "ruondev")==0;
}

static int tryreuse(iao_connection *c) {
	if (doreuse(c) && agentsslsession!=NULL) 
	{
		SSL_set_session(c->ssl, agentsslsession);
		agentsslsession = NULL;
	}
	return 1;
}

static void saveforreuse(iao_connection *c)
{
	if (doreuse(c)) 
	{
		agentsslsession = SSL_get_session(c->ssl);
	}
}

int iao_sslify(iao_connection *result, int milisec)
{
	int stage = 0;
	char d[512];
	result->ssl = SSL_new(sslctx);
	if (result->ssl!=NULL)
	{
		CyaSSL_set_timeout(result->ssl, 24 * 60 * 60);
		stage = 1;
		if (SSL_set_fd(result->ssl, result->fd) && tryreuse(result))
		{
#ifdef WIN32
			int nonblock = 1;
			stage = 2;
			if (ioctlsocket(result->fd, FIONBIO, &nonblock)==0)
#else
			int flags = fcntl(result->fd, F_GETFL);
			int nonblock = flags|O_NONBLOCK;
			stage = 2;
			if (fcntl (result->fd, F_SETFL, nonblock)>=0)
#endif
			{
				int inprogress=1;
				stage = 3;
				while (inprogress)
				{
					int errret;
					int ret = SSL_connect(result->ssl);
					switch (ret)
					{
						case 1:
							saveforreuse(result);
							return 0;
						case 0:
							inprogress = 0;
							stage = 4;
							break;
						default:
							errret = SSL_get_error(result->ssl, ret);
							switch (errret)
							{
								case SSL_ERROR_WANT_READ:
									inprogress = readytorecv(result, milisec)==0;
									stage = 5;
									break;
								case SSL_ERROR_WANT_WRITE:
									inprogress = readytosend(result, milisec)==0;
									stage = 6;
									break;
								default:
									sprintf(d, "iao_sslconnect: SSL_get_error %d %d",
										 ret, errret);
									iao_logmsg(d);
									inprogress = 0;
									stage = 7;
									break;
									
							}
							break;
					}
				}
			}
		}
	}
	sprintf(d, "iao_sslconnect: SSL_connect failed (stage %d)", stage);
	iao_logsysmsg(d);
	result->err = ERR_SSL_FAILED;
	iao_close(result);
	return -1;
}

int iao_sslconnect(char *host, int port, iao_connection *result, int milisec)
{
	int ret = iao_connect(host, port, result);
	if (ret!=0)
	{
		return ret;
	}
	return iao_sslify(result, milisec);
}
#endif

static int udpsend(iao_connection *s, char *buff, int len)
{
	struct sockaddr_in sin;
	int err;
    	sin.sin_family = AF_INET;
    	sin.sin_port = htons(s->remoteport);
    	sin.sin_addr.s_addr = s->udp;
	err = sendto(s->fd, buff, len, 0, (struct sockaddr *) &sin, sizeof(sin));
	if (err==-1)
	{
		s->err = ERR_COMM_DISCONNECTED;
               	iao_logsysmsg("udpsend: sendto returned with an error");
	}
	return err;
}

int iao_send(iao_connection *s, char *buff, int len, int milisec)
{
	if (readytosend(s, milisec)==0) 
	{
		if (s->udp)
		{
			return udpsend(s, buff, len);
		}
#ifdef IAO_SSL
		if (s->ssl==NULL)
		{
#endif
			int err = send(s->fd,buff,len, 0 );
			if (err==-1)
			{
				s->err = ERR_COMM_DISCONNECTED;
				iao_logsysmsg("iao_send: send returned with an error");
				return -1;
			}
			else 
			{
				return err;
			}
#ifdef IAO_SSL
		}
		else
		{
			int err = SSL_write(s->ssl, buff, len);
			if (err<=0)
			{
				iao_logsysmsg("iao_send: SSL_write returned with an error");
				s->err = ERR_COMM_DISCONNECTED;
				return -1;
			}
			else
			{
				return err;
			}
		}
#endif
	}
	else
	{
		return -1;
	}
}

#ifdef IAO_SSL
static int sslrecv(iao_connection *s, char *buff, int len, int milisec)
{
	while (1)
	{
		/* first attempt (in buffer) */
		int err = SSL_read(s->ssl, buff, len); 
		if (err>0) 
		{
			return err;
		}
		else
		{
			if (SSL_get_error(s->ssl, err)==SSL_ERROR_WANT_READ)
			{
				err = readytorecv(s, milisec);
				if (err!=0)
				{
					return -1;
				}
			}
			else
			{
				s->err = ERR_COMM_DISCONNECTED;
				iao_logsysmsg("iao_recv: SSL_read returned with an error");
				return -1;
			}
		}
	}
}
#endif

int udprecv(iao_connection *s, char *buff, int len)
{
	struct sockaddr_in sin;
	unsigned int sinsize = sizeof(sin);
	int ret = recvfrom(s->fd, buff, len, 0, (struct sockaddr *)&sin, &sinsize);
	if (ret<=0)
	{
		s->err = ERR_COMM_DISCONNECTED;
	}
	return ret;
}

int iao_recv(iao_connection *s, char *buff, int len, int milisec)
{
#ifdef IAO_SSL
	if (s->ssl!=NULL)
	{
		return sslrecv(s, buff, len, milisec);
	}
#endif
	if (readytorecv(s, milisec)==0) 
	{
		if (s->udp)
		{
			return udprecv(s, buff, len);
		}
		else
		{
			int err = recv(s->fd,buff,len, 0 );
			if (err<=0)
			{
				s->err = ERR_COMM_DISCONNECTED;
				return -1;
			}
			else 
			{
				return err;
			}
		}
	}
	else
	{
		return -1;
	}
}

void iao_close(iao_connection *s)
{
#ifdef IAO_SSL
	if (s->ssl!=NULL)
	{
		if (SSL_shutdown(s->ssl)==0)
		{
			SSL_shutdown(s->ssl);
		}
		SSL_free(s->ssl);
	}
#endif
#if defined WIN32
	closesocket(s->fd);
#else
	close(s->fd);
#endif
}


int iao_tcpinit()
{
	if (ghbn_mutex==NULL) {
#if defined WIN32
	int ver_mjr=2;
	int ver_mnr=0;
	WORD wVersionRequested=MAKEWORD( ver_mjr, ver_mnr );
	WSADATA wsaData;
	int ret;
	if ((ret=WSAStartup( wVersionRequested, &wsaData ))!=0)
	{
		switch (ret)
		{
		case WSASYSNOTREADY:
			iao_logsysmsg("iao_tcpinit: Underlying network subsystem is not ready");
			break;
		case WSAVERNOTSUPPORTED:
			iao_logsysmsg("iao_tcpinit: WinSock version  not supported");
			break;
		case WSAEINPROGRESS:
			iao_logsysmsg("iao_tcpinit: A blocking Windows Sockets 1.1 operation is in progress");
			break;
		case WSAEPROCLIM:
			iao_logsysmsg("iao_tcpinit: Maximum WinSock tasks reached");
			break;
		case WSAEFAULT:
			iao_logsysmsg("iao_tcpinit: Invalid pointer");
			break;
		default:
			iao_logsysmsg("iao_tcpinit: Unknow reason");
			break;
		}
		return -1;
	}
#else
	/*sigignore(SIGPIPE); changed to a more portable call */
	signal(SIGPIPE, SIG_IGN);
#endif
#ifdef IAO_SSL
	SSL_library_init();
	sslctx = SSL_CTX_new(SSLv3_client_method());
	if (sslctx==NULL)
	{
		iao_logmsg("Failed creating ssl context");
		return -1;
	}
	SSL_CTX_set_verify(sslctx, SSL_VERIFY_NONE, NULL);

	/*
	if ever we want to clean up, call:
     	SSL_CTX_free(sslctx);
`	*/
#endif
	ghbn_mutex = iao_createmutex();
	if (ghbn_mutex==NULL)
	{
		iao_logmsg("Can't create gethostbyname mutex");
		return -1;
	}
	return 0;
} else
	{
		return 0;
	}
}

char *iao_comm_errstr(iao_connection *connection)
{
	static char *ss[] = {
        	"OK",
        	"Local server resource issue",
        	"Can't translate host name",
        	"No connection could be made to host",
        	"Connection actively refused by host",
        	"Timeout occured while negotiating with host",
        	"Host disconnected",
		"SSL negotiation failed"
	};
	return ss[-(connection->err)];
}



int iao_server(int port, iao_connection *result)
{
    struct sockaddr_in sin;
    unsigned int sinlen = sizeof(sin);
 
    resetconnectstruct(result);
    if ((result->fd = (int)socket(AF_INET, SOCK_STREAM,IPPROTO_TCP))<0) 
    {
        result->err = ERR_COMM_RESOURCE;
        iao_logsysmsg("iao_server: error creating socket");
        return -1;
    }

    sin.sin_family = AF_INET;
    sin.sin_port = htons(port);
    sin.sin_addr.s_addr = INADDR_ANY;
    memset(&(sin.sin_zero), 0, sizeof(sin.sin_zero));

    if (bind(result->fd, (struct sockaddr*)&sin, sizeof(sin))!=0)
    {

        iao_logsysmsg("iao_server: bind failed");
        result->err = ERR_COMM_PORTINUSE;
        iao_close(result);
        return -1;
    }

    if (getsockname(result->fd, (struct sockaddr *)&sin, &sinlen)!=0)
    {
        iao_logsysmsg("iao_server: getsockname failed");
        result->err = ERR_COMM_RESOURCE;
        iao_close(result);
        return -1;
    }
    result->myport = ntohs(sin.sin_port);

    if (listen(result->fd, 64)!=0)
    {
        iao_logsysmsg("iao_server: listen failed");
        result->err = ERR_COMM_RESOURCE;
        iao_close(result);
        return -1;
    }
    return 0;
}

int iao_accept(iao_connection *server, iao_connection *client, int milisec)
{
	/* Variables */
	struct timeval time_struct,*tm;
	int err;
    struct sockaddr_in sin;
    unsigned int sockaddrlen = sizeof(sin);


    /* Mask */
	fd_set tmp_Mask, tmp_ExceptMask;
	FD_ZERO(&tmp_Mask);
	FD_SET(server->fd,&tmp_Mask);
	FD_ZERO(&tmp_ExceptMask);
	FD_SET(server->fd,&tmp_ExceptMask);

    resetconnectstruct(client);

    /* time */
	if (milisec==-1)
	{
		tm=NULL;
	}
	else
	{
		tm=&time_struct;
		time_struct=gettimeval(milisec);
	}

	err = select( (unsigned int) server->fd+1,&tmp_Mask,NULL,&tmp_ExceptMask,tm);
    	if (err<0)
	{
		iao_logsysmsg("iao_accept: select returned an error");
		server->err = ERR_COMM_RESOURCE;
		return -1;
	}

	if (FD_ISSET(server->fd,&tmp_ExceptMask))
	{
		iao_logsysmsg("iao_accept: socket closed");
		server->err = ERR_COMM_DISCONNECTED;
		return -1;
	}
	if (!(FD_ISSET(server->fd,&tmp_Mask)))
	{
		server->err = ERR_COMM_TIMEOUT;
		return -1;
	}

    /* Now accept */    
    client->fd = (int) accept(server->fd, (struct sockaddr*) &sin, &sockaddrlen);
    if (client->fd<0)
    {
        iao_logsysmsg("iao_accept: accept failed");
        return -1;
    }
    iptostr(sin.sin_addr.s_addr, client->remotehost);
	return 0;
}

int iao_udp(iao_connection *udp, int myport, char *host, int port) {

	struct sockaddr_in sin;
        unsigned int sinlen = sizeof(sin);
	
	sin.sin_family = AF_INET;
	sin.sin_port = htons(myport);
	sin.sin_addr.s_addr = INADDR_ANY;

	resetconnectstruct(udp);

	udp->udp = iao_gethostbyname(host, udp->remoteip);
	if (udp->udp==0)
	{
		iao_logsysmsg("iao_udp: Can't translate host to ip");
                udp->err=ERR_COMM_HOSTNAME;
		return -1;
	}

	if ((udp->fd = (int) socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP))==-1)
    	{
		iao_logsysmsg("iao_udp: Can't obtain file descriptor");
		udp->err = ERR_COMM_RESOURCE;
		return -1;
    	}
	
	if (bind(udp->fd, (struct sockaddr *) &sin, sizeof(sin))!=0)
	{
		iao_logsysmsg("iao_udp: bind failed");
		udp->err = ERR_COMM_PORTINUSE;
		iao_close(udp);
		return -1;
	}
	if (getsockname(udp->fd, (struct sockaddr *)&sin, &sinlen)!=0)
    	{
        	iao_logsysmsg("iao_udp: getsockname failed");
        	udp->err = ERR_COMM_RESOURCE;
        	iao_close(udp);
        	return -1;
    	}

    	udp->myport = ntohs(sin.sin_port);
	udp->remoteport = port;
	strcpy(udp->remotehost, host);
	return 0;
}



int bridgedata(iao_connection *rcv, fd_set *rcvmask, fd_set *exceptmask, iao_connection *snd, int millisec)
{
	if (FD_ISSET(rcv->fd, exceptmask))
	{
		return -1;
	}
	if (FD_ISSET(rcv->fd, rcvmask))
	{
		char buff[8192];
		int ret = iao_recv(rcv, buff, 8192, millisec);
		if (ret<0) 
		{
			return -1;
		}
		else if (ret>0)
		{
			iao_send(snd, buff, ret, millisec);
		}
	}
	return 0;
}

void iao_bridge(iao_connection *c1, iao_connection *c2, int millisec)
{
	fd_set rcvmask, exceptmask;

	FD_ZERO(&rcvmask);
	FD_ZERO(&exceptmask);

	while (1)
	{	
		int err;
		FD_SET(c1->fd, &rcvmask);
		FD_SET(c2->fd, &rcvmask);
		FD_SET(c1->fd, &exceptmask);
		FD_SET(c2->fd, &exceptmask);
		err = select((unsigned int) (c1->fd>c2->fd?c1->fd:c2->fd)+1,
			 &rcvmask, NULL, &exceptmask, NULL);
		if (err<0)
		{
			iao_logsysmsg("iao_bridge: select returned an error");
			return;
		}
		if (bridgedata(c1, &rcvmask, &exceptmask, c2, millisec) ||
		    bridgedata(c2, &rcvmask, &exceptmask, c1, millisec))
		{
			iao_logmsg("iao_bridge: session disconnected");
			return;
		}
	}
}
/*
** Copyright R-U-ON, 2006
** www.r-u-on.com
*/

#ifdef WIN32
#pragma warning(disable : 4996)
#endif 



#define TIMEOUT  60000

static int iao_get_header(iao_http_channel* channel, char *buff);


static void anerror(iao_http_channel *channel, char *s)
{
	channel->statuscode = -1;
	strcpy(channel->status, s);
}

static void commerror(iao_http_channel *channel)
{
	channel->statuscode = channel->connection.err;
	strcpy(channel->status, iao_comm_errstr(&(channel->connection)));
}

int iao_http_open(char *host, int port, int secure, iao_http_channel *channel)
{
	int ret;
	channel->statuscode = 0;
	channel->status[0]=0;
	channel->proxycridentials[0]=0;
	channel->realhost[0]=0;
	channel->realport = 0;
	channel->contenttype[0] = 0;
	if (secure)
	{
#ifdef IAO_SSL
		ret = iao_sslconnect(host, port, &(channel->connection), TIMEOUT);
#else
		iao_logmsg("iao_http_connect: Can't provide secure connection - not linked with SSL");
		anerror(channel, "SSL not supported by agent");
		return -1;
#endif
	}
	else
	{
		ret =  iao_connect(host, port, &(channel->connection));
	}
	if (ret!=0)
	{
		commerror(channel);
        }

	return ret;
}


#ifdef IAO_SSL
static int proxyconnect(iao_http_channel *channel, char *host, int port, char *proxycridentials)
{
	int ret;
	char buff[10000];
	char *p = buff;
	int len;
	
	p+=sprintf(p, "CONNECT %s:%d HTTP/1.0\r\n", host, port);
        p+=sprintf(p, "User-Agent: ruon/1.0\r\n");


	if (proxycridentials[0]!=0)
	{
		p+=sprintf(p, "Proxy-Authorization: Basic %s\r\n", proxycridentials);
	}
	p+=sprintf(p, "\r\n");
	len = (int)(p - buff);
	ret = iao_send(&(channel->connection), buff, len, TIMEOUT);
	if (ret==len)
	{
		if (iao_get_header(channel, buff)==0)
		{
			if (channel->statuscode==200)
			{
				ret = iao_sslify(&(channel->connection), TIMEOUT);
				if (ret==0)
				{
					return ret;
				}
				commerror(channel);
			}
			else
			{
				sprintf(buff, "proxyconnect: bad http status %d '%s'", 
					channel->statuscode, channel->status);
				iao_logmsg(buff);
				anerror(channel, buff);
			}
		}
		else
		{
			commerror(channel);
			iao_logmsg("proxyconnect: Error receiving header");
		}
	}
	else
	{
		commerror(channel);
		iao_logmsg("proxyconnect: error sending CONNECT request");
	}
	iao_close(&(channel->connection));
	return -1;
}
#endif

int iao_http_proxy_open(char *host, int port, int secure, char *proxyhost, int proxyport, char *proxycridentials, iao_http_channel *channel)
{
	int ret = iao_http_open(proxyhost, proxyport, 0, channel);
	if (ret==0)
	{
		if (secure)
		{
#ifdef IAO_SSL
			return proxyconnect(channel, host, port, proxycridentials);
#else
			iao_logmsg("iao_http_proxy_open: Can't provide secure connection - not linked with SSL");
			anerror(channel, "Can't provide a secure connection - not linked with SSL");
			return -1;
#endif
		}
		else
		{
			if (proxycridentials[0]!=0) 
			{
				strcpy(channel->proxycridentials, proxycridentials);
			}
			strcpy(channel->realhost, host);
			channel->realport = port;
			return 0;
		}
	}
	else
	{
		iao_logmsg("iao_http_proxy_open: Can't connect to proxy");
		commerror(channel);
		return ret;
	}
}


static void httpheader(char *buff, iao_http_channel* channel, char *url, int contentlen)
{
	char *method = contentlen>0?"POST":"GET";

	/* first line */
	if (channel->realhost[0]==0)
        { 
		buff+=sprintf(buff, "%s %s HTTP/1.0\r\n", method, url);
        }
	else
	{
		buff+=sprintf(buff, "%s http://%s:%d%s HTTP/1.0\r\n", method, channel->realhost, channel->realport, url);
	}

	/* User Agent */
	buff+=sprintf(buff, "User-Agent: ruon/1.0\r\n");

	/* Host */
	buff+=sprintf(buff, "Host: %s", channel->connection.remotehost);
	if (channel->connection.remoteport!=80 && channel->connection.remoteport!=443)
	{
		buff+=sprintf(buff, ":%d", channel->connection.remoteport);
	}
	buff+=sprintf(buff, "\r\n");

	/* Keepalive */
	buff+=sprintf(buff, "Connection: Keep-Alive\r\n");

	/* Content Length */
	if (contentlen>0)
	{
		buff+=sprintf(buff, "Content-Length: %d\r\n", contentlen);
	}

	/* Proxy basic authentication */
	if (channel->proxycridentials[0]!=0)
	{
		buff+=sprintf(buff, "Proxy-Authorization: Basic %s\r\n", channel->proxycridentials);
	}

	sprintf(buff, "\r\n");
}

int iao_http_get(iao_http_channel* channel, char *url)
{
	char buff[1024];
	int len;
	int ret;

	httpheader(buff, channel, url, -1);
	len = (int)strlen(buff);

	ret = iao_send(&(channel->connection), buff, len, TIMEOUT);
	if (ret==len) 
	{
		return 0;
	}
	else
	{
		commerror(channel);
		return -1;
	}
}

int iao_http_post(iao_http_channel* channel, char *url, char *data, int datalen)
{
	char buff[1024];
	int len;
	int ret;

	httpheader(buff, channel, url, datalen);
	len = (int)strlen(buff);
	ret = iao_send(&(channel->connection), buff, len, TIMEOUT);
	if (ret!=len) 
	{
		commerror(channel);
		return -1;
	}
	ret=iao_send(&(channel->connection), data, datalen, TIMEOUT);
	if (ret==datalen)
	{
		return 0;
	}
	else
	{
		commerror(channel);
		return -1;
	}
}

static int iao_getbetween(char *start, char *end, char *buff, char *result)
{
	char *es;
	int startlen;
	int resultlen;
	char *ss = strstr(buff, start);
	if (ss==NULL)
	{
		return -1;
	}
	es = strstr(ss, end);
	if (es==NULL) 
	{
		return -1;
	}
	startlen  = (int) strlen(start);
	resultlen = (int) (es-ss-startlen);
	strncpy(result, ss+startlen, resultlen);
	result[resultlen] = 0;
	return 0;
}


static int iao_getattribute(char *attribname, char *buff, char *result)
{
	char b[256];
	strcpy(b, attribname);
	strcat(b, ": ");
	return iao_getbetween(b, "\r\n", buff, result);
}

static int iao_get_header(iao_http_channel* channel, char *buff)
{
	char *eoh = "\r\n\r\n";
	int maxlen=9999;
	int received = 0;
	int eohlen=(int)strlen(eoh);
	int eohrcv=0;
	char b;
	buff[received]=0;
	while (1)
	{
		int ret;
		if (received==maxlen)
		{
			anerror(channel, "Header too big");
			iao_logmsg("iao_get_header: Header too big");
			return -1;
		}
		ret = iao_recv(&(channel->connection), buff+received, 1, TIMEOUT);
		if (ret<=0)
		{
			fflush(stdout);
			iao_logmsg("get_header: disconnected while reading header");
			commerror(channel);
			return -1;
		}
		b = buff[received];
		/*if (b<' ') printf("[%d]", b ); else printf("%c", b);*/
		received++;
		if (received==4 && strncmp("HTTP", buff, 4)!=0) /* not HTTP header */
		{
		 	channel->statuscode = 200;
                        strcpy(channel->status, "OK");
			buff[4] = 0;
			return 0;
		}
		eohrcv = b==eoh[eohrcv] ? eohrcv+1 : 0; /* reset or advance */
		if (eohrcv==eohlen)
		{
			char *sp1, *sp2, *nl;
			char tmp[1024];
			int len;
			char *err = NULL;

			buff[received]=0;
			if ((sp1=strstr(buff, " "))==NULL) { err = "iao_get_header: Can't find space"; }
			if ((sp2=strstr(sp1+1, " "))==NULL) { err = "iao_get_header: Can't find second space"; }
			if ((nl=strstr(buff, "\r\n"))==NULL) { err = "iao_get_header: Can't find newline"; }
			if (err!=NULL)
			{
				iao_logmsg(err);
				anerror(channel ,"Error parsing header");
				return -1;
			}
			len = (int)(sp2-sp1-1);
			strncpy(tmp, sp1+1, len);
			tmp[len]=0;
			channel->statuscode = atoi(tmp);
			len = (int)(nl-sp2-1);
			strncpy(channel->status, sp2+1, len);
			channel->status[len]=0;
			return 0;
		}
	}
}

static int recv2disconnect(iao_http_channel* channel, char **result, char *header)
{
	int alloc = 0;
	int rcvd = 0;
	char b[1024];
	char *buff=NULL;
	int ret;

	if (strncmp(header, "HTTP", 4)!=0) // this is not header but data
	{
		buff = (char*)malloc(10000);	
		alloc = 10000;
		strcpy(buff, header);
		rcvd = (int)strlen(buff);
	}

	do {
		if ((ret = iao_recv(&(channel->connection), b, 1024, TIMEOUT))>0)
		{
			if (ret+rcvd>alloc)
			{
				char *t = (char *) malloc(alloc+10000);
				if (alloc>0)
				{
					memcpy(t, buff, alloc);
					free(buff);
				}
				buff = t;
				alloc+=10000;
			}
			memcpy(buff+rcvd, b, ret);
			rcvd+=ret;
		}
	} while (ret>0);

	if (channel->connection.err!=ERR_COMM_TIMEOUT)
	{
		if (rcvd==0)
                {
                       buff = (char *)malloc(1);
                }

		buff[rcvd]=0;
		*result = buff;
		return rcvd;
	}
	else
	{
		commerror(channel);
		free(buff);
		return -1;
	}
	
}


int iao_http_getpage(iao_http_channel* channel, char **buff)
{
	int received=0;
	int page_size;
	char cl[50];
	char header[10000];

	memset(header, 0, sizeof(header));	

	if (iao_get_header(channel, header)!=0)
	{
		return -1;
	}

	iao_getattribute("Content-Type", header, channel->contenttype);
	if (iao_getattribute("Content-Length", header, cl)!=0)
	{
		return recv2disconnect(channel, buff, header);
	}
	page_size = atoi(cl);

	if (page_size<0)
	{
		anerror(channel, "Negative page size");
		return -1;
	}

	(*buff) = malloc(page_size+1);
	(*buff)[0]=0;
	if (page_size==0)
	{
		return 0;
	}
	while (1)
	{
		int ret = iao_recv(&(channel->connection), (*buff)+received, page_size-received, TIMEOUT);
		if (ret<=-1)
		{
			free(*buff);
			commerror(channel);
			return -1;
		}
		if (ret==0)
		{
			if (received==page_size)
			{
				return received;
			}
			else
			{
				char eb[256];
				free(*buff);
                               sprintf(eb, "Partial data received (%d/%d bytes)", received, page_size); 
				anerror(channel, eb);
				return -1;
			}
		}

		received+=ret;
		(*buff)[received]=0;
		if (received>=page_size)
		{
			return received;
		}
	}
	free(*buff);
	commerror(channel);
	return -1;
}


void iao_http_close(iao_http_channel* channel)
{
	iao_close(&(channel->connection));
}

char *iao_http_status(int code)
{
	typedef struct {
		int code;
		char *s;
	} entry;
	
	static entry codes[] = {
                {100, "Continue"}, {101, "Switching Protocols"}, {200, "OK"},
                {201, "Created"}, {202, "Accepted"}, {203, "Non-Authoritative Information"},
                {204, "No Content"}, {205, "Reset Content"}, {206, "Partial Content"},
                {300, "Multiple Choices"}, {301, "Moved Permanently"}, {302, "Moved Temporarily"},
                {303, "See Other"}, {304, "Not Modified"}, {305, "Use Proxy"},
                {306, "Unused"}, {307, "Temporary Redirect"}, {400, "Bad Request"},
                {401, "Unauthorized"}, {402, "Payment Required"}, {403, "Forbidden"},
                {404, "Not Found"}, {405, "Method Not Allowed"}, {406, "Not Acceptable"},
                {407, "Proxy Authentication Required"}, {408, "Request Time-out"}, {409, "Conflict"},
                {410, "Gone"}, {411, "Length Required"}, {412, "Precondition Failed"},
                {413, "Request Entity Too Large"}, {414, "Request-URI Too Large"}, {415, "Unsupported Media Type"},
                {416, "Requested Range Not Satisfiable"}, {417, "Expectation Failed"}, {500, "Internal Server Error"},
                {501, "Not Implemented"}, {502, "Bad Gateway"}, {503, "Service Unavailable"},
                {504, "Gateway Time-out"}, {505, "HTTP Version not supported"},
		{0,0}
	};

	entry *p = codes;
	while (p->code!=0)
	{
		if (p->code==code)
		{
			return p->s;
		}
		p++;
	}
	
	switch (code/100)
	{
		case 1:
			return "1xx Intermediate status";
		case 2:
			return "2xx Successful response";
		case 3:
			return "3xx Redirect";
		case 4:
			return "4xx Request error";
		case 5:
			return "5xx Server error";
		default:
			return "Unknown HTTP status code";
	}
}



#ifdef WIN32
#pragma warning(disable : 4996)
#endif 

static int geturl(iao_http_channel *channel, char *host, int port, int secure,  char *resource,
		 char **data, char *proxyhost, int proxyport, char *proxycridentials)
{
	int ret;

	if (proxyhost!=NULL)
	{
		ret = iao_http_proxy_open(host, port, secure, proxyhost, proxyport,
                              proxycridentials, channel);
	}
	else
	{
		ret = iao_http_open(host, port, secure, channel);
	}
	if (ret==0)
	{
		if ( (ret = iao_http_get(channel, resource))==0)
		{
			ret = iao_http_getpage(channel, data);
		}
		iao_http_close(channel);
	}
	return ret;
}

static char *parseadvance(char *haystack, char *needle)
{
	int len = (int)strlen(needle);
	if (strncmp(haystack, needle, len)==0)
	{
		return haystack + len;
	}
	else
	{
		return NULL;
	}
}

static void getbetween(char *buff, char *p1, char *p2)
{
	int len = (int)(p2 - p1);
	strncpy(buff, p1, len);
	buff[len]=0;
}

int iao_geturlproxy(iao_http_channel *channel, char *url, char **data, char *proxyhost, 
                    int proxyport, char *proxycridentials)
{
	char *http = "http://";
	char *https = "https://";
	char *p;
	char *colon,*slash;
	char host[1024];
	int port = 80;
	char resource[1024];
	int secure=0;

	p = parseadvance(url, https);
	if (p!=NULL)
	{
		secure = 1;
		port=443;
	}
	else
	{
		p = parseadvance(url, http);
		if (p==NULL)
		{
			p = url;
		}
	}

	colon = strstr(p, ":");
	slash = strstr(p, "/");

	if (colon!=NULL && slash!=NULL && colon>slash)
	{
		colon = NULL;
	}

	if (colon!=NULL)
	{
		port = atoi(colon+1);
	}

	if (slash!=NULL)
	{
		strcpy(resource, slash);
		if (colon!=NULL)
		{
			getbetween(host, p, colon);
		}
		else
		{
			getbetween(host, p, slash);
		}
	}
	else
	{
		strcpy(resource, "/");
		if (colon!=NULL)
		{
			getbetween(host, p, colon);
		}
		else
		{
			strcpy(host, p);
		}
	}

	return geturl(channel, host, port, secure, resource, data, proxyhost, proxyport, proxycridentials);
}

int iao_geturl(iao_http_channel *channel, char *url, char **data)
{
	return iao_geturlproxy(channel, url, data, NULL, -1, NULL);
}
/*
char *agentname() 
{
	return "testagent";
}
int main(int argc, char *argv[])
{
	char *data;
	iao_http_channel http;
	int ret;
	char *url = argc<=1?"http://asdfasdf:23/samba/dibana.html":argv[1];
	
	iao_tcpinit();
	ret = iao_geturl(&http, url, &data);

	printf("\nret = %d\nstatuscode = %d\nstatus = %s\nstatus2 =  %s\n", ret, http.statuscode, http.status, iao_http_status(http.statuscode)); 
	if (ret>0)
	{
		data[100]=0;
		printf("%s\n", data);
		free(data);
	}
	return 0;
}*/
/*
** Copyright R-U-ON, 2006
** www.r-u-on.com
*/


static char cb64[]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
static unsigned char db64[] =
 {62,230,230,230,63,52,53,54,55,56,57,58,59,60,61,230,230,230,230,230,230,230,0,1,
  2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,230,230,230,230,
  230,230,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,
  50,51};

static void encodeblock( unsigned char out[4], unsigned char in[3], int len )
{
    out[0] = cb64[ in[0] >> 2 ];
    out[1] = cb64[ ((in[0] & 0x03) << 4) | ((in[1] & 0xf0) >> 4) ];
    out[2] = (unsigned char) (len > 1 ? cb64[ ((in[1] & 0x0f) << 2) | ((in[2] & 0xc0) >> 6) ] : '=');
    out[3] = (unsigned char) (len > 2 ? cb64[ in[2] & 0x3f ] : '=');
}

static void decodeblock( unsigned char out[3], unsigned char in[4] )
{   
    out[0] = (unsigned char ) (in[0] << 2 | in[1] >> 4);
    out[1] = (unsigned char ) (in[1] << 4 | in[2] >> 2);
    out[2] = (unsigned char ) (((in[2] << 6) & 0xc0) | in[3]);
}


int iao_base64encode(char *dest, void *src, int len) 
{
	int destlen = 0;
	int i;
	for (i=0;i<len;i+=3)
	{
		encodeblock((unsigned char *)dest+destlen, (unsigned char *)src+i, len-i>=3?3:len-i);
		destlen+=4;
		if (destlen%72==0) 
		{
			dest[destlen++]='\n';
		}
	}
	dest[destlen]=0;
	return destlen;
}

int iao_base64decode(void *dest, char *src)
{
	int len = (int) strlen(src);
	int destlen = 0;
	unsigned char block[4];
	int blocklen = 0;
	int i;

	for (i=0;i<len;i++)
	{
		if (src[i] >= 43 && src[i] <=122 && src[i]!='=')
		{
			block[blocklen++] = db64[src[i]-43];
		}
		if (blocklen==4)
		{
			decodeblock( (unsigned char *) dest+destlen, block);
			destlen+=3;
			blocklen=0;
		}
	}
	if (blocklen)
	{
		decodeblock( (unsigned char *) dest+destlen, block);
		destlen+=(blocklen-1);
	}

	return destlen;
}

/*
** Copyright R-U-ON, 2006
** www.r-u-on.com
*/


#ifdef WIN32
#pragma warning(disable:4996)
#endif


typedef struct iao_configpair_struct {
	char name[256];
	char value[1024];
} iao_configpair;
typedef struct iao_configpairs_struct {
	iao_configpair *items;
	int size;
} iao_configpairs;
typedef struct iao_configresource_struct {
	char resourcename[512];
	iao_configpairs pairs;
} iao_configresource;

typedef struct iao_configresources_struct {
	iao_configresource *items;
	int size;
} iao_configresources;
typedef struct iao_config_struct {
	iao_configpairs pairs;
	iao_configresources resources;
} iao_config;

static iao_config agentconfig;
static char *resourcelist = NULL;

static char escapechar(char *s, int len)
{
    typedef struct { char *seq; char chr; int len;} seqt;
    seqt seq[] = {
        {"amp",'&', 3},
        {"quot", '"', 4},
        {"lt", '<', 2},
        {"gt", '>', 2},
        {NULL, 0, 0}};
        seqt *p = seq;

        while (p->seq!=NULL)
        {
            if (p->len==len && strncmp(p->seq, s, len)==0)
            {
                return p->chr;
            }
            p++;
        }
        return 0;
}

static void xmlunescape(char *buff)
{
    char *last = NULL;
    char *p = buff;
    char *c = buff;

    while (*p)
    {
        char chr;
        if (*p==';' && last!=NULL && p-last<6 && 
            (chr = escapechar(last+1, (int)(p-last-1)))!=0)
        {       
            c = c - (p-last);
            *c = chr;
            last = NULL; 
        }       
        else    
        {       
            if (*p=='&')
            {       
                last = p;
            }
            *c = *p;
        }
        c++;
        p++;
    }
    *c = 0;
}


/*static void debugconfig()
{
        int i;
        printf("---START CONFIG---\n");
        for (i = 0;i<agentconfig.pairs.size;++i)
        {
                printf("%s=%s\n", agentconfig.pairs.items[i].name, agentconfig.pairs.items[i].value);
        }
        for (i = 0; i<agentconfig.resources.size;++i)
        {
                int j;
                printf("%s\n", agentconfig.resources.items[i].resourcename);
                for (j = 0;j<agentconfig.resources.items[i].pairs.size;++j)
                {
                        printf("\t%s=%s\n",
                                agentconfig.resources.items[i].pairs.items[j].name,
                                agentconfig.resources.items[i].pairs.items[j].value);
                }
        }
        printf("---END CONFIG---\n");
}*/

static void freepairs(iao_configpairs pairs)
{
	if (pairs.items!=NULL) {
		free(pairs.items);
		pairs.items=NULL;
		pairs.size = 0;
	}
}

static void freeconfig()
{
	int i;

	freepairs(agentconfig.pairs);
	agentconfig.pairs.size = 0;

	if (agentconfig.resources.items!=NULL)
	{
		for (i=0;i<agentconfig.resources.size;++i)
		{
			freepairs(agentconfig.resources.items[i].pairs);
		}
		free(agentconfig.resources.items);
		agentconfig.resources.size = 0;
		agentconfig.resources.items = NULL;
	}
	if (resourcelist!=NULL)
	{
		free(resourcelist);
		resourcelist = NULL;
	}
}

static char *pairsgetvalue(iao_configpairs pairs, char *name)
{
	int i;
	for (i=0;i<pairs.size;++i)
	{
		if (strcmp(name, pairs.items[i].name)==0)
		{
			return pairs.items[i].value;
		}
	}
	return NULL;
}

static int peelnode(char *name, char *buff, int len)
{
	char o[256];
	char c[256];
	int clen;
	int olen;
	
	sprintf(o, "<%s>", name);
	sprintf(c, "</%s>", name);
	olen = (int) strlen(o);
	clen = (int) strlen(c);

	if (memcmp(buff, o, olen)!=0)
	{
		sprintf(c, "peelnode: Can't find opening tag: %s", o);
		iao_logmsg(c);
		return -1;
	}
	if (memcmp(buff+len-clen, c, clen)!=0)
	{
		sprintf(o, "peelnode: Can't find closing tag: %s", c);
		iao_logmsg(o);
		return -1;
	}
	len = len - olen - clen;
	memmove(buff, buff+olen, len);
	*(buff+len)=0;
	return len;
}


static int parsepair(iao_configpair *pair, char *buff)
{
	int len;
	char closingtag[256];
	char *closetag;
	char *openend;

	if (*buff!='<')
	{
		iao_logmsg("parsepair: buff not a node beginning");
		return -1;
	}
	openend = strstr(buff, ">");

	if (openend==NULL)
	{
		iao_logmsg("parsepair: Parse error couldn't find > of opening tag");
		return -1;
	}

	len = (int)(openend - buff -1);
	strncpy(pair->name, buff+1, len);
	pair->name[len]=0;

	sprintf(closingtag, "</%s>", pair->name);
	closetag = strstr(openend+1, closingtag);
	if (closetag==NULL)
	{
		char msg[256];
		sprintf(msg, "parsepair: Can't find closing tag \"%s\"", closingtag);
		iao_logmsg(msg);
		return -1;
	}
	len = (int) (closetag-openend-1);
	strncpy(pair->value, openend+1, len);
	pair->value[len]=0;
	xmlunescape(pair->value);
	return (int)(closetag-buff+strlen(closingtag));
}

static int parsepairs(iao_configpairs *pairs, char *buff, int len)
{
	iao_configpair parsedpairs[256];
	int foundpairs=0;
	int pos=0;

	while (pos<len)
	{
		int ret = parsepair(parsedpairs+(foundpairs++), buff+pos);
		if (ret==-1)
		{
			return -1;
		}
		else
		{
			pos+=ret;
		}
	}

	if (foundpairs==0)
	{
		pairs->items = NULL;
	}
	else
	{
		len = sizeof(iao_configpair)*foundpairs;
		pairs->items = malloc(len);
		memcpy(pairs->items, parsedpairs, len);
	}
	pairs->size = foundpairs;
	return pos;
}

static int parseresource(char *p, iao_configresource *resource)
{
	char *opentag = "<resource>";
	char *closetag = "</resource>";
	char *begin;
	char *end;
	int len;

	begin = strstr(p, opentag);
	end = strstr(p, closetag);
	if (begin==NULL || end==NULL)
	{
		iao_logmsg("parseresource: Couldn't find <resource></resource> pair");
		return -1;
	}

	len = parsepairs(&resource->pairs, begin+strlen(opentag), (int)(end-begin-strlen(opentag)) );
	if (len==-1)
	{
		iao_logmsg("parseresource: Failed parsing resource");
		return -1;
	}

	if (resource->pairs.size>0)
	{
		strcpy(resource->resourcename, resource->pairs.items[0].value);
	}
		
	return (int)(end-begin+strlen(closetag));
}

static int parseresources(char *p)
{
	iao_configresource resources[256];
	int parsedresources=0;

	if (peelnode("resources", p, (int)strlen(p))==-1)
	{
		return -1;
	}

	while (*p)
	{
		int len = parseresource(p, &resources[parsedresources++]);
		if (len==-1)
		{
			return -1;
		}
		p+=len;
	}

	if (parsedresources==0)
	{
		agentconfig.resources.items=NULL;
	}
	else
	{
		int len = parsedresources*sizeof(iao_configresource);
		agentconfig.resources.items=malloc(len);
		memcpy(agentconfig.resources.items, resources, len);
	}
	agentconfig.resources.size = parsedresources;
	return 0;
}

static void buildresourcelist()
{
	int len = 0;
	char *p;
	int i;
	for (i=0;i<agentconfig.resources.size;++i)
	{
		len+=(int)(strlen(agentconfig.resources.items[i].resourcename)+1);
	}
	if (len==0)
	{
		return;
	}

	resourcelist = malloc(len+1);
	p = resourcelist;
	for (i=0;i<agentconfig.resources.size;++i)
	{
		int len = sprintf(p, "%s",
                             agentconfig.resources.items[i].resourcename);
		p+=(len+1);
	}
	*p = 0;
}

int iao_parseconfig(char *buff, int len)
{
	char *p;

	freeconfig();

	len = peelnode("config", buff, len);
	if (len==-1)
	{
		return -1;
	}

	/* Find end of pairs */
	p = strstr(buff, "<resources>");
	
	len = p==NULL?len:(int)(p-buff);
	
	len = parsepairs(&agentconfig.pairs, buff, len);
	if (len==-1) 
	{
		return -1;
	}
	if (p!=NULL)
	{
		if (parseresources(p)==-1)
		{
			return -1;
		}
		buildresourcelist();
	}
	/*debugconfig();*/
	return 0;
}

char *iao_configvalue(char *name)
{
	return pairsgetvalue(agentconfig.pairs, name);
}


char *iao_configresourcevalue(char *resource, char *name)
{
	int i;
	for (i=0;i<agentconfig.resources.size;++i)
	{
		if (strcmp(resource, agentconfig.resources.items[i].resourcename)==0)
		{
			return pairsgetvalue(agentconfig.resources.items[i].pairs, name);
		}
	}
	return NULL;
}

char *iao_configresourcelist()
{
	return resourcelist;
}

void iao_configinit()
{
	agentconfig.resources.items = NULL;
	agentconfig.resources.size = 0;
	agentconfig.pairs.items = NULL;
	agentconfig.pairs.size = 0;
	resourcelist = NULL;
}
/*
** Copyright R-U-ON, 2006
** www.r-u-on.com
*/



#ifdef WIN32
#pragma warning(disable : 4996)
#endif 

#define IAOP_TAG_OPEN "<iaop>"
#define IAOP_TAG_CLOSE "</iaop>"

#define IAOP_VERSION 1
#define IAOP_SERVLET "/iaop1"

static time_t iao_born_time;
static int iao_iteration = 0;
static char host[IAOP_MAX_LEN];
static int port;
static char proxyhost[IAOP_MAX_LEN];
static int proxyport;
static char proxycridentials[IAOP_MAX_LEN];
static int doproxy = 0;
static iao_mutex mx;

static int iao_firsttime()
{
	static int first_api_call=1;
	if (first_api_call==1)
	{
		char buff[256];
		iao_born_time = time(NULL);
		if (iao_tcpinit()!=0)
		{
			return -1;
		}
		mx = iao_createmutex();
		if (mx==NULL)
		{
			iao_logmsg("Can't create main mutex");
			return -1;
		}
		iao_configinit();

		if (iao_getembeddedvariable("RUON_HOST", host)==-1)
		{
#ifdef CIAO_TEST
			strcpy(host, "ruondev");
#else
			strcpy(host, "agent.r-u-on.com");
#endif
		}

		if (iao_getembeddedvariable("RUON_PORT", buff)==0)
		{
			port = atoi(buff);
		}
		else
		{
#ifdef IAO_SSL
#ifdef CIAO_TEST
			port = 8443;
#else
			port = 443;
#endif
#else
			port = 80;
#endif
		}

		if (iao_getembeddedvariable("PROXY_HOST", proxyhost)==0)
		{
			iao_logmsg("Using proxy settings");
			doproxy = 1;
			iao_getembeddedvariable("PROXY_PORT", buff);
			proxyport = atoi(buff);
			proxycridentials[0]=0;
			iao_getembeddedvariable("PROXY_CRIDENTIALS", proxycridentials);
		}

		first_api_call=0;
		return 0;

	}
	return 0;
}

int iao_init()
{
	return iao_firsttime();
}

static void iao_datetime(char *buff)
{
	time_t t = time(NULL);
	struct tm * ts = localtime(&t);
	sprintf(buff, "%04d-%02d-%02d %02d:%02d:%02d", 
		ts->tm_year+1900, ts->tm_mon+1, ts->tm_mday,
		ts->tm_hour, ts->tm_min, ts->tm_sec);
}

static void iao_buildpostbuff(iao_agentinfo *agent, iao_http_channel *http, char *buff, char *request)
{
	char td[128];
	iao_datetime(td);

	sprintf(buff, "<iaop pver=\"%d\" atype=\"%s\" aver=\"%s\" os=\"%s\" id=\"%s\" ip=\"%s\" host=\"%s\""
                      " time=\"%s\" uptime=\"%d\" iter=\"%d\">%s</iaop>", 
		IAOP_VERSION,
		agent->type,
		agent->version,
		OS,
		agent->id,
		http->connection.myip,
		http->connection.myhost,
		td,
		(int)(time(NULL)-iao_born_time),
		iao_iteration,
		request
		);
}

static int iao_peelnode(char *name, char *buff, int len)
{
	char o[256];
	char c[256];
	int clen;
	int olen;
	
	sprintf(o, "<%s>", name);
	sprintf(c, "</%s>", name);
	olen = (int) strlen(o);
	clen = (int) strlen(c);

	if (memcmp(buff, o, olen)!=0)
	{
		sprintf(c, "iao_peelnode: Can't find opening tag: %s", o);
		iao_logmsg(c);
		return -1;
	}
	if (memcmp(buff+len-clen, c, clen)!=0)
	{
		sprintf(o, "iao_peelnode: Can't find closing tag: %s", c);
		iao_logmsg(o);
		return -1;
	}
	len = len - olen - clen;
	memmove(buff, buff+olen, len);
	*(buff+len)=0;
	return len;
}

static void iao_error(char *p)
{
	char buff[1024];
	iao_peelnode("error", p, (int)strlen(p));
	sprintf(buff, "iao_error: server returned an error: %s", p);
	iao_logmsg(buff);
}

static char *mylastip=NULL;
static void remembermyip(iao_http_channel *http)
{
	static char ip1[20];
	static char ip2[20];

	char *ip = http->connection.myip;
	if (mylastip==NULL)
	{
		strcpy(ip1, ip);
		mylastip = ip1;
	}
	else
	{
		char *newplace = mylastip==ip1?ip2:ip1;
		strcpy(newplace, ip);
		mylastip = newplace;
	}
}
char *iao_getmylastip()
{
	return mylastip;
}

static int iao_post_blocked(iao_agentinfo *agent, char *request, char **buff)
{
	int len;
	char post_buff[10240];
	iao_http_channel http;
	char *die = "<directive>die:";
	static int consecutive_not_iaop_compliant = 0;
#ifdef IAO_SSL
	int secure = 1;
#else
	int secure = 0;
#endif

	if (iao_firsttime()!=0)
	{
		return -1;
	}

	if (doproxy==1)
	{
		if (iao_http_proxy_open(host, port, secure, proxyhost, 
                    proxyport, proxycridentials, &http)!=0)
		{
			return -1;
		}
	}
	else 
	{
		if (iao_http_open(host, port, secure, &http)!=0) 
		{
			return -1;
		}
	}

	remembermyip(&http);
	

	iao_buildpostbuff(agent, &http, post_buff, request);

	if (iao_http_post(&http, IAOP_SERVLET, post_buff, (int)strlen(post_buff))!=0)
	{
		iao_http_close(&http);
		return -1;
	}

	len = iao_http_getpage(&http, buff);
	iao_http_close(&http);

	if (len<0) 
	{
		return -1;
	}

        len = iao_peelnode("iaop", *buff, len);
        if (len==-1)
        {
                iao_logmsg(*buff);
                iao_logmsg("iao_post: response not IAOP compliant");
                free(*buff);
		if ((consecutive_not_iaop_compliant++)==2000)
		{	
			iao_logmsg("iao_post: response not IAOP compliant for 200 times in a row");
			iao_logmsg("iao_post: There could be a firewall issue.");
			iao_logmsg("iao_post: Bailing out.");
			exit(-1);
		}
                return -1;
        }
	consecutive_not_iaop_compliant = 0;

	if (strncmp(*buff, "<error>", 7)==0)
	{
		iao_error(*buff);
		free(*buff);
		return -1;
	}

	if (strncmp(*buff, die, strlen(die))==0)
	{
		iao_logmsg(*buff);
		iao_logmsg("iao_post: Was asked to die by server. Bailing out...");
		free(*buff);
		exit(-1);
	}
	return len;
}

static int iao_post(iao_agentinfo *agent, char *request, char **buff)
{
	int ret;

	iao_lockmutex(mx);
	ret = iao_post_blocked(agent, request, buff);
	iao_releasemutex(mx);
	return ret;
}

static void nextiteration()
{
	iao_lockmutex(mx);
	iao_iteration++;
	iao_releasemutex(mx);
}

int iao_register(iao_agentinfo *agent, char *result_agent_id, int skaa) 
{
	char *buff;
	int len;
	char *method = skaa?"<register skaa=\"true\"/>":"<register/>";

	len = iao_post(agent, method, &buff);
	if (len<0)
	{
		return -1;
	}

	if (strncmp("<id>", buff, 4)!=0)
	{
		iao_logmsg("iao_register: failed!");
		iao_logmsg(buff);
		free(buff);
		return -1;
	}

	strncpy(result_agent_id, buff+4, len-9);
	result_agent_id[len-9]=0;
	free(buff);
	nextiteration();
	return 0;
}

static int directive(iao_agentinfo *agent, char *inner, char *verb, char *x)
{
	char *buff;
    int len = iao_post(agent, inner, &buff);
	char *del;

    if (len<0)
    {
            return -1;
    }

    len = iao_peelnode("directive", buff, len);
    if (len<0)
    {
            free(buff);
            return -1;
    }

    del = strstr(buff, ":");
    if (del==NULL)
    {
            iao_logmsg("directive: malformed directive, no ':'");
            free(buff);
            return -1;
    }       
            
    *del=0;
    strcpy(verb, buff);
    if (strcmp(verb,"ERROR")==0)
    {
            iao_logmsg("directive: Got an ERROR from the server");
            iao_logmsg(del+1);
            return -1;
    }
    strcpy(x, del+1);
    free(buff);
	nextiteration();
    return 0;
}

int iao_keepalive(iao_agentinfo *agent, char *verb, char *x)
{
	return directive(agent, "<keepalive />", verb, x);
}

static char *addbuff(char *p, char *buff)
{
	strcat(p, buff);
	return p+strlen(buff);
}

static char *addalarm(char *p, iao_alarm *alarm)
{
	char *closetag;
	switch (alarm->type)
	{
	case 'C':
	case 'c':
		alarm->type = 'C';
		p = addbuff(p, "<clear ");
		closetag = "</clear>";
		break;
	case 'E':
	case 'e':
	case 'A':
	case 'a':
	default:
		if (alarm->type=='e'||alarm->type=='E')
		{
			p = addbuff(p, "<event ");
			closetag = "</event>";
		}
		else
		{
			p = addbuff(p, "<alarm ");
			closetag = "</alarm>";
		}
		switch(alarm->severity)
		{
		case 'M':
			p = addbuff(p, "severity=\"M\" ");
			break;
		case 'C':
			p = addbuff(p, "severity=\"C\" ");
			break;
		case 'm':
		default:
			p = addbuff(p, "severity=\"m\" ");
			break;
		}
		break;
	}

	p = addbuff(p, "resource=\"");
	p = addbuff(p, alarm->resource);
	p = addbuff(p, "\" id=\"");
	p = addbuff(p, alarm->id);
	p = addbuff(p, "\">");
	if (!(alarm->description[0]==0 && alarm->type=='C'))
	{
		p = addbuff(p, "<![CDATA[");
		p = addbuff(p, alarm->description);
		p = addbuff(p, "]]>");
	}
	p = addbuff(p, closetag);
	return p;
}

static void validatexml(char *buff)
{
	static char **transform=NULL;
	char *p = buff;
	if (transform==NULL)
	{
		transform = malloc(256*sizeof(char*));
		memset(transform, 0, 256*sizeof(char*));
		transform['&'] = "&amp;";
		transform['<'] = "&lt;";
		transform['\''] = "&apos;";
		transform['"'] = "&quot;";
		transform['>'] = "&gt;";
	}

	while (*p!=0)
	{
		if (transform[(int)(*p)]!=NULL)
		{
			char *tt = transform[(int)(*p)];
			int tlen = (int)strlen(tt);

			memmove(p+tlen, p+1, strlen(p));
			memcpy(p, tt, tlen);
			p+=(tlen-1);
		}
		p++;
	}
}

static void alarmslist(char *buff, iao_alarm alarms[], int count)
{
	char *p = buff+strlen(buff);
	int i;
	for (i=0;i<count;++i)
	{
		validatexml(alarms[i].id);
		validatexml(alarms[i].resource);
		p = addalarm(p, &alarms[i]);
	}
}

int iao_alarms(iao_agentinfo *agent, iao_alarm alarms[], int count,
               int inc, char *verb, char *x) 
{
	char alist[10240];

	iao_logmsg("iao_alarms: reporting agent state");
	sprintf(alist, "<alarms incremental=\"%s\">", inc?"true":"false");
	alarmslist(alist, alarms, count);
	strcat(alist, "</alarms>");

	return directive(agent, alist, verb, x);
}

int iao_alarms_ifchange(iao_agentinfo *agent, iao_alarm alarms[], int count,char *verb, char *x)
{
#define MAX_ALARM_BUFF_SIZE 50000
	static char *lasttime = NULL;
	static time_t lasttimetime = 0;
	time_t now = time(NULL);
	int effectivecount = 0;
	char thistime[MAX_ALARM_BUFF_SIZE]="";
	char *p = thistime;
	int i;

	for (i=0;i<count;++i)
	{
		if (alarms[i].type!='C')
		{
			p+=sprintf(p, "%s", alarms[i].id);
			p+=sprintf(p, "%s", alarms[i].resource);
			effectivecount++;
			if (p-thistime>MAX_ALARM_BUFF_SIZE-1000)
			{
				iao_logmsg("iao_alarms_ifchange: too many alarms");
				return iao_alarms(agent, alarms, count, 0, verb, x);
			}
		}
	}
	if (lasttime==NULL)
	{
		lasttime = (char *)malloc(MAX_ALARM_BUFF_SIZE);
		lasttime[0]=0;
	}
	if (strcmp(lasttime, thistime)==0 && (now-lasttimetime<60*60))
	{
		return 1;
	}
	else
	{
		int ret = iao_alarms(agent, alarms, count, 0, verb, x);
		if (ret==0)
		{
			strcpy(lasttime, thistime);
			lasttimetime = time(NULL);
		}
		return ret;
	}
}

static int buildreqvar(char *str, iao_confparam *param)
{
        return sprintf(str, "<param name=\"%s\" type=\"%s\" default=\"%s\"/>",
                param->name, param->type, param->defvalue);
}
static int buildreqvars(char *tag, char *str, iao_confparam *params)
{
        char *p = str;
        int i=0;

        p+=sprintf(p, "<%s>",tag);
        while (params+i!=NULL)
        {
		if ( (params+i)->name==NULL)
		{
			break;
		}
                p+=buildreqvar(p, params+i);
                i++;
        }
        p+=sprintf(p, "</%s>",tag);
        return (int) (p-str); 
}       

int iao_fetchconfig(iao_agentinfo *agent,
                    iao_confparam *agentconf, iao_confparam *resources)
{
        char buff[8192];
        char *p = buff;
	char *retbuff;
	int len;
        
        p+=sprintf(p, "<fetchconfig>");
        if (agentconf!=NULL)
        {
                p+=buildreqvars("agent", p, agentconf);
        }
        if (resources!=NULL)
        {
                p+=buildreqvars("resources", p, resources);
        }
        sprintf(p, "</fetchconfig>"); 

        len = iao_post(agent, buff, &retbuff);
        if (len<0)
        {
                return -1;
        }
        
        if (strncmp("<config>", retbuff, 8)!=0)
        {       
                iao_logmsg("iao_fetchconfig: failed!");
                iao_logmsg(retbuff);
                free(retbuff);
                return -1;
        }

		iao_lockmutex(mx);
		iao_parseconfig(retbuff, len);
		iao_releasemutex(mx);

        free(retbuff);
        return 0;

}




#define EVNC "!!!EMBEDDED_VARIABLE_NOT_CONFIGURED!!!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ENDOFBUFF"

static char *iao_getembededbuff()
{
	static char *embedded_variables = EVNC;
	return embedded_variables;
}

int iao_getembeddedvariable(char *name, char *result) 
{
	char *nstart;
	char *vstart;
	char *vend;
	size_t len;

	char nnn[IAOP_MAX_LEN];

	sprintf(nnn, "<NAME/>%s<VALUE/>", name);
	nstart = strstr(iao_getembededbuff(), nnn);
	if (nstart==NULL) 
	{
		return -1;
	}
	vstart = nstart+strlen(nnn);
	vend = strstr(vstart, "<END/>");
	if (vend==NULL)
	{
		return -1;
	}
	len = vend - vstart;
	strncpy(result, vstart, len);
	result[len] = 0;
	return 0;
}

static int advance(char *token, char **buff)
{
	int len = (int)strlen(token);
	if (strncmp(token, *buff, len)==0) 
	{
		(*buff) = *buff + len;
		return 0;
	}
	else
	{
		return -1;
	}
}

static void *mymemmem(void *big, int biglen, void *sub, int sublen)
{
	int i;
	for (i=0;i<biglen-sublen;i++)
	{
		if (memcmp( ((char*)big)+i, sub, sublen)==0)
		{
			return ((char*)big)+i;
		}
	}
	return NULL;
}

static void proxiate(char *buff, int len)
{
	char ph[IAOP_MAX_LEN];
	char pp[IAOP_MAX_LEN];

	if (iao_getembeddedvariable("PROXY_HOST", ph)==0)
	{
		char *srch = "<END/>@@@@@@@@@@@@@@@@@@@@@";
		int slen = (int)strlen(srch);
		char *found = mymemmem(buff, len, srch, slen);
		iao_getembeddedvariable("PROXY_PORT", pp);

		if (found!=NULL)
		{
			char addbuff[512];
			int addlen = sprintf(addbuff, "<NAME/>PROXY_HOST<VALUE/>%s<END/>"
			                              "<NAME/>PROXY_PORT<VALUE/>%s<END/>", ph, pp);
			found+=strlen("<END/>");
			memcpy(found, addbuff, addlen);
		}
		iao_getembeddedvariable("PROXY_PORT", pp);

		
		
	}
}


int iao_getbinary(iao_agentinfo *agent, void **buff, int *size)
{
	char authenticbuff[IAOP_MAX_LEN];
	char *authentic=NULL;
	int len;
	char *head1 = "<binary authentic=\"";
	char *head2 = "\">";
	char *tail = "</binary>";
	int tailen;
	char *p;
	void *bin;

	if (iao_getembeddedvariable("AUTHENTIC_TAG", authenticbuff)==0) 
	{
		authentic = authenticbuff;
	}
	else
	{
		iao_logmsg("iao_getbinary: AUTHENIC_TAG variable not embedded");
	}

	len = iao_post(agent, "<getbinary />", (char**)buff);
	if (len<0)
	{
		return -1;
	}

	p = *buff;

	if (advance(head1, &p)!=0)
	{
		iao_logmsg("iao_getbinary: <binary> tag invalid");
		free(*buff);
		return -1;
	}
	if (authentic)
	{
		if (advance(authentic, &p)!=0)
		{
			iao_logmsg("iao_getbinary: binary not authentic");
			free(*buff);
			exit(-1);
		}
	}
	else
	{
		p = strstr(p, head2);
	}

	if (advance(head2, &p)!=0)
	{
                iao_logmsg("iao_getbinary: <binary> tag  invalid (closing)");
                free(*buff);
                return -1;
	}

	tailen = (int) strlen(tail);
	if (strncmp(tail, ((char*)*buff)+len-tailen, tailen)!=0) 
	{
                iao_logmsg("iao_getbinary: </binary> tag invalid");
                free(*buff);
                return -1;
	}

	*(((char*)*buff)+len-tailen) = 0; /* terminate b64 with a null */

	bin = malloc(len);
	*size = iao_base64decode(bin, p);
	proxiate(bin, *size);
	
	free(*buff);
	*buff = bin;
	return 0;
}
 
static int isstandardparam(char *name) 
{
	return strcmp(name, IAO_AGENTPARAM_ALIAS)==0 ||
	       strcmp(name, IAO_AGENTPARAM_GROUP)==0 ||
	       strcmp(name, IAO_AGENTPARAM_NOTIF)==0 ||
	       strcmp(name, IAO_AGENTPARAM_SKAA)==0 ||
	       strcmp(name, IAO_AGENTPARAM_RESOURCES)==0;
}

int iao_setparams(iao_agentinfo *agent, iao_param namevalues[]) 
{
	char buffer[10000];
	char *p = buffer;
	iao_param *param = namevalues;
	int len;

	p+=sprintf(p, "<setparam>");

	while (param->name) 
	{
		char *isconfig = 
			isstandardparam(param->name) ?
			"":"config.";

		p+=sprintf(p, "<%s%s><![CDATA[%s]]></%s%s>",
			isconfig, param->name, param->value,
			isconfig, param->name);
		param++;
	}
	p+=sprintf(p, "</setparam>");

        len = iao_post(agent, buffer, &p);
        if (len<0)
        {
                return -1;
        }
	free(p);
	return 0;
}

